"""
Installation Engine

  an abstract engine, that performs the
  key tasks of resolution of dependencies and
  provide a framework for executing tasks that
  will install dependencies.

  key operations

   - resolve all dependencies given current installation

   - empty db resolve all dependencies

  key concepts to information discovery
 
   - package database, container of installed packages

   - package repository, container of available packages

   - package metadata of package to be installed

  key concepts to dependency resolution
  
   - packages structured as a set of nodes

   - nodes organized into a DAG

   - version numbers as distutils.version.StrictVersion

   - versions organized into axis of package node

   - edges as node, version, operator

   - edges not resolved to runtime, to allow for context
     dependent edges
     
  Copyright
    
     @2002 kapil thangavelu <kvthan@wm.edu>
     
  License
  
     GPL
     
  CVS: $Id: InstallEngine.py,v 1.2 2002/10/01 10:55:42 ender Exp $
  
  Thoughts: 
      
      - still need to figure out db/repo interaction
      - which is equiv to integration of version axis.
"""


import operator
from distutils.version import StrictVersion
from unittest import TestCase, TestSuite, main, makeSuite

from Pools import GraphTaskPool
from Task import ITask, Task
from Utils import Interface, Attribute

#################################
# Exceptions

class InvalidVersionOperator(Exception):
    pass

class InvalidDependencyTask(Exception):
    pass

class DependencyVersionConflict(Exception):
    pass

#################################
# Interfaces

class IDependencyVersion(Interface):
    """
    represents a particular version of a dependency,
    provides for version comparison.
    """

class IDependency(Interface):
    """
    A dependency represents a directed edge between a node
    and another node. 
    """

    def getId():
        """ 
        return id of the dependency, id should
        correspond to a task name
        """

    def getOperator():
        """
        return the operator function to use in comparison
        of dependencies
        """
        
    def getVersion():
        """
        return the version that the dependency points to
        """
        
class IDependencyTask(ITask):

    def getDependencies(execution_context=None):
        """
        returns a list of dependencies
        (objects not node_ids).

        depending on settings, some dependencies
        may be dependent on execution_context.
        """

#################################
# the interfaces are the following two are
# subject to extreme change.
#
# they currently are minimal interfaces
# intended to be mixed into a more
# complete interface representing
# package database, and package repositories
# respectively.

class IDependencyDatabase(Interface):

    def getDependencyTask(dependency):
        """
        return a dependency task with
        a version axis corresponding
        to all installed versions. if
        no version is installed return
        None.
        """

class IDependencyRepository(Interface):

    def getDependency(dependency):
        """
        return a dependency task with
        a version axis corresponding
        to all available versions, if
        no versions exist return None.
        """

#################################
# Implementation

class DependencyVersion(StrictVersion):
    __implements__ = IDependencyVersion

class VersionAxis:

    def __init__(self, versions):

        self.versions = []
        for v in versions:
            if not IDependencyVersion.isImplemntedBy(v):
                v = DependencyVersion(v)
            self.versions.append(DependencyVersion(v))
        self.versions.sort()

    def getMatches(self, version, operator):

        assert IDependencyVersion.isImplemntedBy(version)
        op_func = getOperator(operator)
        return filter(lambda v, tv=version, op=op_func: op(v, tv), self.versions)


class Dependency:

    __implements__ = IDependency
    
    def __init__(self, id, version, operator='greater-equal' ):

        self.id = id
        self.version = StrictVersion(version)
        if isOperator(operator):
            self.operator = operator
        else:
            raise InvalidVersionOperator("Not a valid operator %s"%str(op))            

    def getId(self):
        return self.id
        
    def getVersion(self):
        return self.version
    
    def getOperator(self):
        return getOperator(self.operator)
    
    def __str__(self):
        return "%s %s %s %s"%(self.__class__.__name__,
                              self.id,
                              str(self.version),
                              str(self.operator))
                           
class DependencyTask(Task):

    __implements__ = IDependencyTask

    def getDependencies(self, execution_context=None):
        return tuple(self.depends)

class DependencyGraph:

    """
    another incomplete and hairbrained dag
    implementation.

    version conflicts are fatal.
    """

    def __init__(self):
        self._depends = {}
        self._edges = {}
        self._nodes = {}

    def addNode(self, node):

        node_id = node.getId()

        if self._nodes.has_key(node_id):
            node = isVersionCompatible(node, self.getNode(node_id))

        depends = node.getEdges()
        
        for d in depends:
            self._depends.setdefault(d, []).append(node_id)

        self._edges[node_id] = depends
        self._nodes[node_id] = node
        
    def getAdjacent(self, node_id):
        
        return map(self.getNode, self._edges.get(node_id, ()))

    def getTransitive(self, node_id, node_set=None):

        if node_set is None:
            node_set = [node_id]
         
        for d in self._edges.get(node_id, ()):
            nid = d.getId()
            node_set.append(nid)
            self.getTransitive(nid, node_set)
            
        return node_set

    def getNode(self, node_id):
        return self._nodes[node_id]
    
    def getNodeIds(self):
        return self._nodes.keys()
        
                 
class DependencyGraphPool(GraphTaskPool):
    
    def __init__(self, id, priority=1):
        self.id = id
        self.data = DependencyGraph()
        self.priority = priority

#################################
# utility functions

VERSION_OPERATORS = {
    '>': operator.__gt__,
    '>=':operator.__ge__,
    '=':operator.__eq__,
    '<':operator.__lt__,
    '<=':operator.__le__
    }

def isOperator(op):
    global VERSION_OPERATORS
    return VERSION_OPERATORS.has_key(op)

def getOperator(op):
    global VERSION_OPERATORS
    op_func =  VERSION_OPERATORS.get(op, None)
    if op_func is None:
        raise InvalidVersionOperator("Not a valid operator %s"%str(op))
    return op_func

def isVersionCompatible(t1, t2):
    """
    given two tasks of the same name,
    determine if they are version 
    compatible based on their dependencies
    taking into account dependency operators 
    either returns the dtask to use or raises
    an error.
    """
    
    d1 = {}
    d2 = {}
    
    add = lambda d, dstore=d1: operator.__setitem__(dstore, d.getId(), d)
    map(add, t1.getDependencies())
    add = lambda d, dstore=d2: operator.__setitem__(dstore, d.getId(), d)    
    map(add, t2.getDependencies())

    common = filter(d2.has_key, d1.keys())
    
    # really this could be a failing of the model we're using
    # in that we can't really say one dtask is a more recent
    # version and should have precendence over another, because
    # we don't attach a version to the dtask, just a v. axis.
    
    if len(d1) == 0:
        if len(d2) != 0:
            return t2
        return t1 # for all intents and purposes they are equal

    # so they both have non zero dependencies... if they aren't
    # all the same dependencies we got problems...
    if not len(common) == len(d1) and not len(common) == len(d2):
        raise DependencyVersionConflict("%s Mismatch %s\n\n and\n\n %s"%(
                                                    t1.getId(),                            
                                                    str(d1.values()),
                                                    str(d2.values())
                                                    )
                                            )                                                                           
                                        
    for c in common:
        c1 = d1[c]
        c2 = d2[c]
        discrim = cmp(c1.version, c2.version)
        
        # versions equal
        if discrim == 0:            
            
            # check if the same
            if hash(str(c1)) == hash(str(c2)):
                return t2
            # check compatible operators            
            elif c1.getOperator()(c2.version,c1.version) \
               and c2.getOperator()(c1.version, c2.version):
                # maybe need some sort of precendence to operators...
                return t2
            # if not... print for now
            else: 
                print 'warninging minor version conflict'
                return t2
            
        # c1 is bigger
        elif discrim == 1:
            temp = c2
            c2 = c1
            c1 = temp
            del temp
            
        # ok on to operator checks
        
        if not c1.getOperator()(c2.version, c1.version):
            #import pdb; pdb.set_trace()
            
            #print 'hello', 1
           
            raise DependencyVersionConflict("%s Conflict %s and %s"%(
                                                    t1.getId(),                            
                                                    str(c1),
                                                    str(c2)
                                                    )
                                            )
        
        # we should probably leave this as a user tunable setting
        # basically if the older package satisifies the dependency
        # we leave it in place.
        if c2.getOperator()(c1.version, c2.version):
            return t1
        else:
            return t2
                                                                             
        
