"""

Contexts are basically dictionaries used
for value lookup. attribute lookup is allowed.

Context Chains are a stack of dictionaries
that are searched in order.

The Context Stack is a global stack of
dictionaries. it offers methods to create
a chain from the stack composed of arbitrary
elements.

The Context Registry, is a registry of
chain names to context name sets, its used
internally by the context stack to resolve
chain_name -> chain

these could really benefit from using python22isms

Author: kapil thangavelu <kvthan@wm.edu>
License: GPL
Copyright: kapil thangavelu @2002
CVS: $Id: Context.py,v 1.4 2002/10/23 01:44:50 ender Exp $
"""

import os
from unittest import TestCase, TestSuite, main, makeSuite

from UserDict import UserDict
from Utils import Interface
from Exceptions import ContextLookupError

_marker = []

class IContext(Interface):
    pass

class IContextChain(Interface):
    pass

class IContextStack(Interface):
    pass

class IContextChainRegistry(Interface):
    pass
    

class Context(UserDict):

    __implements__ = IContext

    def __init__(self, id, data={}):
        self.id = id
        UserDict.__init__(self, data)

    def getId(self):
        return self.id
    
    def __getattr__(self, key):

        if self.data.has_key(key):
            return self.data[key]
        raise ContextLookupError(
            " %s not found in Context %s"%(str(key), self.getId())
            )

class ContextChain(Context):

    __implements__ = IContextChain

    def __init__(self, id, context_names, contexts):
        
        self.context_names = list(context_names)
        self.contexts = list(contexts) # list of contexts
        Context.__init__(self, id)
        
    def getContext(self, name):
        return self.contexts[self.context_names.index(name)]

    def __getattr__(self, name):

        v = _marker

        v = self.data.get(name, _marker)
        if v is not _marker:
            return v
        
        for c in self.contexts:
            v = c.get(name, _marker)
            if v is not _marker:
                break
        
        if v is _marker:
            raise ContextLookupError(
                " %s not found in ContextChain %s"%(str(name), self.getId())

                )
        return v


    def get(self, name, default=_marker):
        v = self.data.get(name, _marker)
        if v is not _marker:
            return v
        
        for c in self.contexts:
            v=c.get(name, default)
            if v is not default:
                break

        if v is not default:
            return v
        
        if default is not _marker:
            return default

        raise ContextLookupError(
            " %s not found in ContextChain %s"%(str(name), self.getId())
            )

    def has_key(self, name):
        _m = []
        if self.get(name, _m) is _m:
            return 0
        return 1

    def keys(self):
        keys = []
        keys.extend(self.data.keys())
        for c in self.contexts:
            keys.extend(c.keys())
        return keys

    debug = keys
    def __getitem__(self, name):
        return self.get(name)
        
    def push(self, context):

        if not IContext.isImplementedBy(context):
            raise ContextError(
                "push arg does not implement icontext %s"%str(context)
                )
        
        self.context_names.insert(0,context.getId())
        self.contexts.insert(0,context)

    def pop(self, n=0):
        self.context_names.pop(n)
        return self.contexts.pop(n)

    def popContext(self, context_name):
        
        idx = self.context_names.index(context_name)
        return self.pop(idx)

    
class ContextStack(ContextChain):
    
    __implements__ = IContextStack

    def __init__(self, id):
        self.id = id
        self.context_names = []
        self.contexts = []
        # name to index... as i expect the most frequent
        # op, is getting chains rather than push/pop
        # self.context_map = {}
        # expect maintaining it is a pain
        
    def getContextChain(self, names, explicit_order=1):
        i = self.context_names.index
        g = lambda n, c=self.contexts: c[n]
        idxs = map(i, names)
        contexts = map(g, idxs)

        if not explicit_order:
            slist = zip(idxs, contexts)
            slist.sort()
            names = [self.context_names[e[0]] for e in slist]
            contexts = [e[1] for e in slist]

        # anonymous chain
        return ContextChain('anonymous-chain', names, contexts)
    
    def getContextChainByName(self, chain_name):
        return self.getContextChain(getContextsFor(chain_name), 1)

    __getattr__ = None
    __setitem__ = None
    __getitem__ = None

class _ContextChainRegistry:

    # needs alot more error handling/checking

    __implements__ = IContextChainRegistry
        
    def __init__(self):
        self._chains = {} # chain name -> contexts
        self._contexts = {} # contexts -> chains

    def defineContextChain(self, chain_name, context_names):
        assert isinstance(chain_name, type(''))
        assert isinstance(context_names, type([]))

        self._chains[chain_name]=context_names

        for c in context_names:
            self._contexts.setdefault(c, []).append(chain_name)
    
    def getContextsFor(self, chain_name):
        return self._chains[chain_name]

    def getChainsFor(self, context_name):
        return self._contexts[context_name]


ContextChainRegistry = _ContextChainRegistry()
del _ContextChainRegistry
defineContextChain = ContextChainRegistry.defineContextChain
getContextsFor = ContextChainRegistry.getContextsFor
getChainsFor   = ContextChainRegistry.getChainsFor

stack = ContextStack('ContextStack')
def getContext():
    return stack

