import unittest

from Exceptions import InvalidTask
from Log import getLog
from Structure.Graph import Graph
from Structure.Set import Set
from Structure.Tree import Tree
from Task import ITask

from Utils import unique, Interface, Attribute, SimpleClassRegistry, \
     default_identifier

class ITaskPool(Interface):

    def getId():
        """
        return pool identity
        """

    def addTask(task):
        """
        add a task to the pool, task should
        implement ITask interface
        """
        
    def getTask(task_id):
        """
        retrieve the task with the given id,
        raises an error if the task is not found
        """
    
    def getTaskList(execution_context):
        """
        get a list of task ids, in the order
        in which they should be run.
        """

    def runTasks():
        """
        execute the tasks in the proper order. requires
        a runner to be set.
        """

    def setupEnvironment():
        """
        if the pool environment requires any setup before
        execution of tasks, this method should take
        care of it.
        """

    def tearDownEnvironment():
        """
        for pool cleanup after task execution
        """

    def shouldRun(context):
        """
        return boolean on whether this pool should have
        its tasks run
        """

    def getPriority():
        """
        priority of this pool
        """

    def setTaskRunner(runner):
        """
        set the runner for executing this pools tasks. 
        """

    def getTaskRunner():
        """
        return the task runner associated with this
        pool or None
        """

    priority = Attribute("priority")

class AbstractDataStructTaskPool:

    __implements__ = ITaskPool

    def __init__(self, id, priority):
        raise NotImplemented(" %s is an abstract class"%str(self.__class__))
        
    def addTask(self, task):

        if not ITask.isImplementedBy(task):
            raise InvalidTask(
                "object does not implement task interface %s"%str(task))            

        self.data.addNode(task)

    def getId(self):
        return self.id

    def getTask(self, task_id):
        return self.data.getNode(task_id)
        
    def getTaskList(self, execution_context):
        raise NotImplemented(" %s is an abstract class"%str(self.__class__))

    def setTaskRunner(self, runner):
        self.runner = runner

    def getTaskRunner(self):
        return getattr(self, 'runner', None)
    
    def runTasks(self, execution_context):
        tasks = [self.getTask(t) for t in self.getTaskList(execution_context)]
        self.getTaskRunner().executeTasks(execution_context,tasks)

    def setupEnvironment(self):
        pass
    
    def tearDownEnvironment(self):
        pass

    def shouldRun(self, context=None):
        return 1
                    
class GraphTaskPool(AbstractDataStructTaskPool):

    def __init__(self, id, priority=1):
        self.id = id
        self.data = Graph()
        self.priority = priority

    def getTaskList(self, execution_context):

        res   = []
        taskids = self.data.getNodeIds()
        
        for t in taskids:
            res.extend(self.getDependenciesForTask(t))
        return unique(res)

    def getDependenciesForTask(self, task_id):
        task_set = self.data.getTransitive(task_id)
        task_set.reverse()
        return unique(task_set)

class SetTaskPool(AbstractDataStructTaskPool):

    def __init__(self, id, priority=1):
        self.id = id
        self.data = Set()
        self.priority = priority

    def getTaskList(self, execution_context):
        return unique(self.data.getNodeIds())
        
    def __len__(self):
        return len(self.data)

class TreeTaskPool(AbstractDataStructTaskPool):

    def __init__(self, id, priority=1):
        self.id = id
        self.data = Tree()
        self.priority = priority

    def getTaskList(self, execution_context):
        root_nodes = self.data.getRootNodes()
        

PoolRegistry = SimpleClassRegistry('PoolRegistry')
PoolRegistry.setIdentifier(default_identifier)
registerPool = PoolRegistry.addClass
getPool = PoolRegistry.getClass

registerPool(GraphTaskPool, 'Graph')
registerPool(SetTaskPool,   'Set')

