from Utils import Interface, Attribute, SimpleClassRegistry
from Context import Context
from UserList import UserList
from Log import getLog
from Exceptions import ConfigurationError

from os import path

import glob
from ConfigParser import ConfigParser

class IReader(Interface):
    """
    once a reader reads a source
    it is an iterable set.

    there is a pair wise association
    between readers and sources
    """

    format = Attribute("Format that the reader processes")
    
    def read(context):
        """
        read the configuration, transforming to
        a set of contexts for potential tasks
        at a minimum each task must define the framework
        reserved words:

          - Module
          - Task

          optionally
          
          - Depends
          - ObjectByTaskId
          - Function
          - Class
          
        """

    def __getitem__(i):
        """
        after a reader has had its read method
        called, it should act as an iterator
        """
    
class Reader(UserList):

    __implements__ = IReader

    def read(self, context):
        raise NotImplemented

    def setContext(self):
        raise NotImplemented

    def __str__(self):
        return '%s'%str(self.__class__)

class INIFileReader(Reader):

    def read(self, context):

        log = getLog(self)
        
        handle = open(context.Source)
        parser = ConfigParser()
        parser.readfp(handle)
        handle.close()

        log.debug('file reader found %d contexts', len(parser.sections()))
        self.data.extend( config2Context([], parser) )

    def setContext(self, context):
        pass

class INIDirectoryReader(Reader):

    def __init__(self):
        self.id = 'INIDirectoryReader'
        UserList.__init__(self)
        
    def getId(self):
        return self.id
    
    def read(self, context):

        log = getLog(self)

        directory = context.SourceDirectory
        log.debug('loading files from %s'%directory)
        
        files   = glob.glob1(directory, context.TaskFilePattern)
        log.debug('found %d matching config files'%len(files))

        if not files or not directory:
            log.warn('no files found in config directory %s', directory )
            return []

        files = map(path.join, [directory]*len(files), files)
        parser = ConfigParser()
        
        for f in files:
            handle = open(f)
            parser.readfp(handle)
            handle.close()

        log.debug('converting configs 2 contexts')
        self.data.extend(config2Context([], parser))

def reader_identity(klass):
    return klass.format

ReaderRegistry = SimpleClassRegistry()
ReaderRegistry.setIdentifier(reader_identity)
registerReader = ReaderRegistry.addClass
getReader = ReaderRegistry.getClass

registerReader(INIFileReader, 'ini')
registerReader(INIDirectoryReader, 'ini-dir')


def config2Context(set, parser):

    add = set.append
    entities = parser.sections()

    meta = {'module'  : None,
            'task'    : None,
            'class'   : None,
            'depends' : None }

    for e in entities:
        if not e.count('Task') > 0:
            continue
            
        d = {}
        options = parser.options(e)
        for o in options:
            d[o]=parser.get(e,o)

        module = d.get('module', None)
        function = d.get('task', None)
        depends = d.get('depends', None)
        
        # XXX backwards compat hack
        if function is None:
            function = d.get('function', None)
            if function is not None:
                d['task']=function
        klass = d.get('class', None)
        

        if module is None:
            raise ConfigurationError(" %s No Module Specified"%e)
        if klass and function:
            raise ConfigurationError(" %s Both Class and Function Specified"%e)
        if not klass and not function:
            raise ConfigurationError(" %s No Class or Function Specified"%e)                
        if depends is not None:
            depends = map(lambda x: x.strip(), depends.split(','))
        else:
            depends = ()
        d['depends']=depends
        
        add(Context(e, d))
        
    return set
