import inspect
from lib.mapply import mapply

from Utils import Interface, SimpleClassRegistry
from Exceptions import FunctionSpecError, TaskException

class INode(Interface):

    def getId():
        """ get the node id  """

    def getEdges():
        """ get nodes that leave from this node """

class ITask(INode):

    def getDependencies():
        """
        returns a tuple of task ids that this task
        is dependant on. equivalent to INode.getEdges
        """

    def getContext():
        """
        return the context associated with this
        tasks definition.
        """

    def getType():
        """
        return the type of task
        """

    def __call__(context):
        """
        perform the action related to this task, context
        is the execution context for this task. 
        """

class BaseTask:

    __implements__ = ITask
    
    def getId(self):
        return self.id

    def getDependencies(self):
        return tuple(self.depends)

    def setContext(self, context):
        self.context=context
        
    def getContext(self):
        c = getattr(self, 'context', None)
        if c is None:
            raise TaskException( " no context set %s"%self.getId())
        return c

    def getType(self):
        return self.type
        
    getEdges = getDependencies

    def __str__(self):
        return "<%s> %s"%(self.__class__.__name__, self.getId())

    __repr__ = __str__
    
#################################
# for instance based tasks
class Task(BaseTask):

    def __init__(self, id, depends=()):
        self.id = id
        self.depends = depends

    def __call__(self, context):
        raise NotImplementedError("Task __call__ not implemented")

    def getTaskType(self):
        return self.type

class ZopeTask(Task):

    __implements__ = IZopeTask

    def __call__(self, configuration, persistent_object):
        raise NotImplementedError("ZopeTask __call__ not implemented")

class PoolTask(Task):

    type = 'Pool'

    def __init__(self, pool):
        self.id = pool.getId()
        self.pool = pool
        
    def __call__(self, context):

        if self.pool.shouldRun():
            self.pool.setupEnvironment()
            self.pool.runTasks(context)
            self.pool.tearDownEnvironment()

        return None

#################################
# for function based tasks

class FunctionTask(BaseTask):

    type = 'Function'
    
    __implements__ = ITask
    
    def __init__(self, id, task, depends=()):
        self.id = id
        self.depends = depends
        self.func = task

    def __call__(self, context):

        return apply(self.func, (context,))


class MapplyFunctionTask(BaseTask):

    type = 'mapply'
    
    def __init__(self, id, task, depends=()):
        self.id = id
        self.depends = depends
        self.func = task
            
    def __call__(self, context):
        return mapply(self.func,
                      keyword=context
                      )

TaskRegistry = SimpleClassRegistry()
registerTask = TaskRegistry.addClass
getTask = TaskRegistry.getClass

registerTask(FunctionTask, 'Function')
registerTask(MapplyFunctionTask, 'Engine')

